/* bench - driver for Plum Hall benchmarks 
 * Thomas Plum, Plum Hall Inc,  tplum@plumhall.com
 *
 * Written in "Typesafe C", the common subset of ANSI C and C++
 * Copyright (C) 1991-2002, Plum Hall Inc.
 * Permission granted to reproduce for any purpose,
 * provided this copyright notice is preserved in the copy
 */
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
/* put the following line in a separate source file,  f3.c ...
   void f3() { }
 */
void tabulate(void (*fn)(int, char*[]), char *s);
void benchreg(int ac, char *av[]);
void benchsho(int ac, char *av[]);
void benchlng(int ac, char *av[]);
void benchdbl(int ac, char *av[]);
void benchfn(int ac, char *av[]);

int verbose=1;

int main(int argc, char *argv[])
	{
	char result[5][10];
	int i;

	if (argc < 3)
		{
		fprintf(stderr, "usage: bench2 1 'compiler-id'\n");
		exit(2);
		}
	if (argv[1][0] != '1')
		printf("argv[1] must be   1   !\n");
	tabulate(benchreg, result[0]);
	tabulate(benchsho, result[1]);
	tabulate(benchlng, result[2]);
	tabulate(benchfn,  result[3]);
	tabulate(benchdbl, result[4]);
	printf("Plum Hall benchmark results, in nanoseconds:\n\n");
	printf("%20.20s  %9s %9s %9s %9s %9s\n",
		"", "register", "auto", "auto", "function", "auto");
	printf("%20.20s  %9s %9s %9s %9s %9s\n",
		"", "int", "short", "long", "call+ret", "double");
	printf("%22.22s ",
		argv[2]);
	for (i = 0; i <= 4; ++i)
		printf("%9.9s ", result[i]);
	printf("\n");
	return 0;
	}
void tabulate(void (*fn)(int, char*[]), char *s)
	{
	static char arg1[20];
	static char *arga[3] = { "x", &arg1[0], 0 };
	double before, after, nanosec;
	double resolution;
	long major, major_next;

	before = (double)clock();
	do	{
		after = (double)clock();
		} while (before == after);
	resolution = after - before;
	major_next = 1;
	do  {
		major = major_next;
		sprintf(arg1, "%ld", major);
		before = (double)clock();
		(*fn)(2, arga);
		after = (double)clock();
		major_next *= 10;
		} while (after-before < 250*resolution);
	nanosec = 1e6 * (after - before) / CLOCKS_PER_SEC / major;
	sprintf(s, "%8.3g ", nanosec);
	if (verbose)
		{
		printf("resolution=%12.8g\n",resolution);
		printf("after     =%12.8g\n",after);
		printf("before    =%12.8g\n",before);
		printf("diff      =%12.8g\n",after-before);
		printf("nanosec   =%12.8g\n",nanosec);
		printf("s         =%s\n",s);
		}
	}

/* benchreg - benchmark for  register  integers 
 * If machine traps overflow, use an  unsigned  type 
 * Let  T  be the execution time in milliseconds
 * Then  average time per operator  =  T/major  usec
 * (Because the inner loop has exactly 1000 operations)
 */
void benchreg(int ac, char *av[])
	{
	register int a, b, c;
	long d, major;
	static int m[10] = {0};

	major = atol(av[1]);
	printf("executing %ld iterations\n", major);
	a = b = (av[1][0] - '0');
	for (d = 1; d <= major; ++d)
		/* inner loop executes 1000 selected operations */
		for (c = 1; c <= 40; ++c)
			{
			a = a+b+c; b = a>>1; a = b%10; m[a] = a; b = m[a]-b-c;
			a = b==c; b = a|c; a = !b; b = a+c; a = b>c;
			}
	printf("a=%d\n", a);
	}
/* benchsho - benchmark for  auto short  integers
 * If machine traps overflow, use an  unsigned  type 
 */
void benchsho(int ac, char *av[])
	{
	auto short a, b, c;
	long d, major;
	static short m[10] = {0};

	major = atol(av[1]);
	printf("executing %ld iterations\n", major);
	a = b = (av[1][0] - '0');
	for (d = 1; d <= major; ++d)
		/* inner loop executes 1000 selected operations */
		for (c = 1; c <= 40; ++c)
			{
			a = a+b+c; b = a>>1; a = b%10; m[a] = a; b = m[a]-b-c;
			a = b==c; b = a|c; a = !b; b = a+c; a = b>c;
			}
	printf("a=%d\n", a);
	}
/* benchlng - benchmark for  auto long  integers
 * If machine traps overflow, use an  unsigned  type 
 */
void benchlng(int ac, char *av[])
	{
	auto long a, b, c;
	long d, major;
	static long m[10] = {0};

	major = atol(av[1]);
	printf("executing %ld iterations\n", major);
	a = b = (av[1][0] - '0');
	for (d = 1; d <= major; ++d)
		/* inner loop executes 1000 selected operations */
		for (c = 1; c <= 40; ++c)
			{
			a = a+b+c; b = a>>1; a = b%10; m[a] = a; b = m[a]-b-c;
			a = b==c; b = a|c; a = !b; b = a+c; a = b>c;
			}
	printf("a=%d\n", a);
	}
/* benchdbl - benchmark for  double   
 */
void benchdbl(int ac, char *av[])
	{
	auto double a, b, c;
	long d, major;

	major = atol(av[1]);
	printf("executing %ld iterations\n", major);
	a = b = (av[1][0] - '0');
	for (d = 1; d <= major; ++d)
		/* inner loop executes 1000 selected operations */
		for (c = 1; c <= 40; ++c)
			{
			a = a+b+c; b = a*2; a = b/10; a = -a; b = -a-b-c;
			a = b==c; b = a+c; a = !b; b = a+c; a = b>c;
			}
	printf("a=%d\n", a);
	}
/* benchfn - benchmark for function calls
 */
int dummy = 0;

extern void f3();  /* in separate source file */

void f2() { f3();f3();f3();f3();f3();f3();f3();f3();f3();f3();} /* 10 */
void f1() { f2();f2();f2();f2();f2();f2();f2();f2();f2();f2();} /* 10 */
void f0() { f1();f1();f1();f1();f1();f1();f1();f1();f1();} /* 9 */

void benchfn(int ac, char *av[])
        {
        long d, major;

        major = atol(av[1]);
        printf("executing %ld iterations\n", major);
        for (d = 1; d <= major; ++d)
                f0(); /* executes 1000 calls */
        printf("dummy=%d\n", dummy);
        }

