/* This is the example movie rendering module for the C/C++ Users Journal
   Report article: A 3D Stereo Movie Viewer with GLUT, Sept 2003

   To run the exuectable, type
 
      cujmovie<ENTER>

   If your hardware supports quad buffered stereo, start with 

      cujmovie -q<ENTER>
   
   
   This demo simply displays 9 rotating 3D objects, and responds to 
   user inputs and mouse clicks which control coloring, center 
   of rotation, etc.
   
   I.e., main() calls dispuiMain() which is in dispui.c.  From there,
   dispui responds to GLUT callbacks, and calls the display* functions
   below to perform movie-specific rendering.
   
   By tailoring the routines below, your movie can leverage the power of
   the dispui.c module.  You can focus on rendering your movie frames, and
   other movie-specific features.  dispui.c will handle the dirty work
   of window management, translation, rotation, scaling, and stereo
   
   Please cite the C/C++ Users Journal article 
      A 3D Stereo Movie Viewer with GLUT
      Sept 2003
      Chris Moth
   in any derivatives of this work.

 *  Current contact information:
 *
 *    Chris Moth
 *    chris.moth@vanderbilt.edu
 *    Phone: 615-936-3569
 *    http://www.structbio.vanderbilt.edu/~cmoth
 *
 *  A much more complex rendering example can be found at the above URL
 *  Follow the links for MDDisplay V3.0
*/ 

// Some useful ANSI C standard headers
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <assert.h>
#include <errno.h>
#include <time.h>

// Include the GLUT libraries and OpenGL headers.
#ifdef __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif

// Include function declarations for the dispui.c module.
#include "dispui.h"

// Prototypes for functions declared here

#include "cujmovie.h"

// The "movie" here is 9 spinning glut* 3D objects.

// Let's set some initial colors for each object
int objectColors[9] = {nWhite,nMagenta,nWhite,nBlue,nAqua,nGreen,nYellow,nRed,nGray};

// When the user clicks on the "static" menu, we'll first draw a greyed
// static image for comparison.  The default is to not do this.
int statictype = nStaticNone;  // nStaticFirst, nStaticLast, nStaticNone

// Define the size of the coordinate system that will be manged when
// scaling is 1.0
const int movieImageSize=100;

// We need to place the first object in the center of the movie window.
// We'll place the others in the 8 corners of the scene.
#define tLoc 30.0

static const float objectCoord[9][3] =
{{0.0,0.0,0.0},         // Object 1 in center
 {-tLoc, tLoc,-tLoc},   // Object 2 in front, top, left
 {tLoc,tLoc,-tLoc},     // 3 in front, top right
 {-tLoc,-tLoc,-tLoc},   // 4 in front bottom left
 {tLoc,-tLoc,-tLoc},    // 5 in front bottom right
 {-tLoc, tLoc,tLoc},    // Object 6 in back, top, left
 {tLoc,tLoc,tLoc},      // 7 in back, top right
 {-tLoc,-tLoc,tLoc},    // 8 in back bottom left
 {tLoc,-tLoc,tLoc}};    // 9 in back bottom right


// dispui.c call will call the functions below to 
// get the first and last frame numbers.
int displayMovieMinFrame(void)
{
   return 0; // Shows as "Frame 1"
}

int displayMovieMaxFrame(void)
{
   return 179;
}

// When the user clicks on an object, we toggle it's rendering
// from wire to solid.  Store the state of each object in this 
// array
static int renderAsSolid[9]; // 1=Solid, 0=Wire

// We render the objects not only when dispui needs a frame update.
// but also when the user is picking with the mouse (gl_mode == GL_SELECT)
// For this simple movie, I've put all cases in a simple shared routine below

void cujRenderFrame(int frame,      // ranges from 1 to 180
                    int gl_mode,    // Will be either GL_RENDER or GL_SELECT
                    int static_flag,// True if we are putting a grey backround image down
                    int smearFlag// True if we are displaying a "smeared frame"  
                    )
{
   // At each new frame
   float angle = 2.0 * (frame);
   
   // Let's position and display the 9 objects
   int whichObject;

   for (whichObject = 0; whichObject < 9;whichObject++)
      {
      char str[4];
      
      // Save the rotation/translation/scaling context that dispui has prepared for us
      glPushMatrix();
      
      // If we are picking with the mouse, then let OpenGL know the label for what we
      // are rendering.
      if (gl_mode == GL_SELECT)
         glLoadName(whichObject+1);
         
      // It's easiest to think of OpenGL transformations as last to first
      // Lastly, place the object in the right cornder (or center) of the movie
      glTranslatef(objectCoord[whichObject][0],
                    objectCoord[whichObject][1],
                    objectCoord[whichObject][2]);
                    
      // First rotate the object.
      glRotatef(angle,0.0,1.0,0.0);
      
      if (! static_flag) 
         {
         // This is the real frame - color last selected color
         dispuiColorByPalette(objectColors[whichObject]);
         }
      else
         {
         // Dull grey for this background static image
         glColor3ub(90,90,90); 
         }
         
      // With the object positioned and color set above,
      // we are ready to draw the object.
      if (renderAsSolid[whichObject])
         {
         switch (whichObject)
            {
            case 0: glutSolidSphere(14.0,12.0,12.0);
                    break;
            case 1: glutSolidCube(24.0);
                    break;
            case 2: glutSolidTorus(6.0,9.0,20.0,20.0);
                    break;
            case 3: glScalef(13.0,13.0,13.0);
                    glutSolidIcosahedron();
                    break;
            case 4: glScalef(12.0,12.0,12.0);
                    glutSolidOctahedron();
                    break;
            case 5: glScalef(15.0,15.0,15.0);
                    glutSolidTetrahedron();
                    break;
            case 6: glScalef(7.0,7.0,7.0);
                    glutSolidDodecahedron();
                    break;
            case 7: glutSolidCone(12.0,17.0,5.0,6.0);
                    break;
            case 8: glutSolidTeapot(15.0);
                    break;
            }
         }
      else
         {
         switch (whichObject)
            {
            case 0: glutWireSphere(14.0,12.0,12.0);
                    break;
            case 1: glutWireCube(24.0);
                    break;
            case 2: glutWireTorus(6.0,9.0,20.0,20.0);
                    break;
            case 3: glScalef(13.0,13.0,13.0);
                    glutWireIcosahedron();
                    break;
            case 4: glScalef(12.0,12.0,12.0);
                    glutWireOctahedron();
                    break;
            case 5: glScalef(15.0,15.0,15.0);
                    glutWireTetrahedron();
                    break;
            case 6: glScalef(7.0,7.0,7.0);
                    glutWireDodecahedron();
                    break;
            case 7: glutWireCone(12.0,17.0,5.0,6.0);
                    break;
            case 8: glutWireTeapot(15.0);
                    break;
            }
         }
      // Return the coordinate transformation back to what we started with - so
      // that next object can be simply rotated and placed.
      glPopMatrix();
      
      // Put an identifying digit "below" each object
      if ((gl_mode == GL_RENDER) && (! smearFlag))
         {
         glColor3ub(255,255,255);
         sprintf(str,"%d",whichObject+1);
         glRasterPos3i(objectCoord[whichObject][0],
                    objectCoord[whichObject][1]-18,
                    objectCoord[whichObject][2]);

         dispuiDrawStringAtCurrentRasterPosition(GLUT_BITMAP_HELVETICA_18,str);
         }

      }

   // Frame all the objects in a  cube
   if(! smearFlag)
      {
      glColor3ub(255,255,255);
      glutWireCube(movieImageSize);
      }

}

// dispui.c will call this function when it needs to draw
// the scene
void displayRenderFrameInPerspectiveBox(int frame,int smearFlag)
{
   if (statictype == nStaticFirst) // Render dull grey "first" frame
      cujRenderFrame(1,GL_RENDER,1,smearFlag);
      
   if (statictype == nStaticAverage) // Render dull grey "avg" frame
      cujRenderFrame(90,GL_RENDER,1,smearFlag);
   
   cujRenderFrame(frame,GL_RENDER,0,smearFlag);
}

void displayRenderFrameInOrthoBox(int frame,int movieWindowWidth,int movieWindowHeight)
{
   // After you have rendered your scene in Perspective, dispui will call this
   // function to render in an Ortho box.  This is particluarly useful if you
   // want to show some text information in the movie window at fixed positions.
   // For the CUJ demo, we do nothing.
}

// dispui.c will call this function when it gets a mouse click
// and needs a list of objects near the mouse click.
void displayRenderFrameInSelectMode(int frame)
{
   cujRenderFrame(frame,GL_SELECT,0,0);
}

// dispui.c will call this function when the clipping
// control is active.  In theory you could save
// your main window as an openGL displayList or use
// other tricks to speed up rendering of the clipping window.
// Here, I jsut draw the whole scene again.
void displayRenderClipToolFrame(int frame)
{
   cujRenderFrame(frame,GL_RENDER,0,0);
}

// Picking 3D objects with a 2D mouse cursor is a bit tricky - but OpenGL gives us
// some powerful help.
//
// When the user clicks in the main Window, dispui calls displayRenderFrameInSelectMode
// which labels each object to OpenGL using the glLoadName function.  
// Then, OpenGL makes a list of "hits" in the selectBuf array.  Finally, dispui
// calls the displayProcessHits function below to let us determine the closest "hit" and
// deal with that in an application-specific manner.
// 
// For this CUJ demo:
//   If coloring is being input, then the clicked object is colored.
//   If center-of-rotation is being input, then then the new c-o-r is the clicked object
//   If no input is active, then the object is toggled between wire frame and solid rendering

static int lastPickedObject = 0;                // Use this to highlight whatever object we clicked on

// If color or center-of-rotation input is taking place, then this function will point to
// a function to deal with that.
static void (* ObjectPickedFunc)(void);

void displayProcessHits(GLint hits,             // The number of objects "near" our mouse click
                        const GLuint* selectBuf // 4 GLuints for each hit
                        )
{
   lastPickedObject = 0; // By default, nothing has been selected.

   if (hits > 0)
      {
      int i;
      
      // Assume that the first "hit" is the best we have.
      const GLuint* hitRecord = &selectBuf[0];
      unsigned int bestZ = hitRecord[1]; // distance from user
      lastPickedObject = hitRecord[3];   // Object ID from glLoadName

      // If there are more than 1 hit - consider all the others.
      // If any are referring to objects which are closer to the user in a 3D sense,
      // then those should be given priority.
      for (i=1;i<hits;i++)
         {
         const GLuint* hitRecord = &selectBuf[i*4];

         // If this hit is "closer" to the viewer then it is better
         if (hitRecord[1] < bestZ) 
            {
            bestZ = hitRecord[1];  // minZValue
            lastPickedObject = hitRecord[3]; // Object ID From glLoadName
            }
         }
         
      if (ObjectPickedFunc != 0)
         {
         // If we have an active "handler" function for dealing 
         // with atoms as they are picked, then call away!
         // This will be the case when the user is coloring objects
         // or selecting a new center of rotation
         ObjectPickedFunc();
         }
      else 
         {
         // Otherwise simply change from solid to wire
         renderAsSolid[lastPickedObject-1] = ! renderAsSolid[lastPickedObject-1];
         }

      MovieAdvanceTimerKeyAndPostRedisplay();
      } // end hits > 0

   //  for debug: printf("%d was lastPicked\n",lastPickedObject);
}


// ** The Following functions deal with text input from users ** 

// When the user has pressed RETURN or ESC during text input of an object
// number to color, this function is called by dispui

static int newColorJustRequested;
void NewColorInputComplete(int lastKey,const char* inputString)
{
   int whichObject;
   // Did user end input with ENTER and enter _something_?  
   // If not (likely was ESCape), ignore
   if ((lastKey == 13) && inputString[0])
      {
      whichObject = atoi(inputString) - 1; // Get the input number
      
      // make sure the entered number makes sense
      // and then color the object
      if (whichObject >= 0 && whichObject <= 9)
         objectColors[whichObject] = newColorJustRequested;
      else
         {
         // Bad input - so put a status message on the movie Window
         sprintf(dispuiStatusMessage,"Invalid Input - no object %-.20s",inputString);
         dispuiStatusMessageExpiration = time(NULL) + 4;
         }

      MovieAdvanceTimerKeyAndPostRedisplay();
      }
   // Now that text input is complete, the text box will go away.
   // Don't color picked objects anymore.
   ObjectPickedFunc = 0;
}

// When coloring, ObjectPickedFunc points to the function below.
// It is called from displayProcessHits above.  Picked objects
// are colored here - and the scene is refreshed.
void NewColorObjectPicked(void)
{
   objectColors[lastPickedObject-1] = newColorJustRequested;
   MovieAdvanceTimerKeyAndPostRedisplay();
}

// Called from dispui.c when a color palette selection has 
// been made or keystroke (0-9) selected.
void displayNewColorRequested(int _newColor)
{
   static char prompt_string[100];

   // Let's store the color selected by the user in a static global
   newColorJustRequested = _newColor;
   
   // Let's ask the user which object he wants to color
   sprintf(prompt_string,"Enter object number (1-9) or pick object to color %s:",
               dispuiPaletteColorString(_newColor));

   // Call the dispui Keyboard input function, passing the address of the 
   // NewColorInputComplete function to be called when the user is done
   // with input.
   KeyboardInputStart(prompt_string,
                     NewColorInputComplete);

   // And, if the user picks an object we want that to be immediately colored.
   ObjectPickedFunc = NewColorObjectPicked;
}

// When the user presses <ENTER> or <ESC> at the end of
// the "center of rotation prompt" dispui calls this funciton
void CenterOfRotationInputComplete(int lastKey,const char* inputString)
{
  // Did user end input with a CR?  
  // If not, it was ESC - so ignore input
  if (lastKey == 13)
      {
      int whichObject = atoi(inputString) -1 ;
      // For a new center of rotation to make sense, we need
      // to reset any translation requested previously
      // by the user
      dispuiTranslation(0.0,0.0);
      if (whichObject < 0) // Then no input entered...
         {
         // We'll restore center of rotation to the center
         // of the coordinate system.
         dispuiCenterOfRotation(0.0,0.0,0.0);
         }
      else
         {
         // If a particular object number was requested, 
         // then make it center of rotation.
         if (whichObject >= 0 && whichObject < 9)
            {
            dispuiCenterOfRotation(objectCoord[whichObject][0],
                        objectCoord[whichObject][1],
                     objectCoord[whichObject][2]);
            }
         else
            {
            // Otherwise, the user entered a bogus number.
            // Put up a status message.
            sprintf(dispuiStatusMessage,"Invalid Input - no object %-.20s",inputString);
            dispuiStatusMessageExpiration = time(NULL) + 4;
            }
         }
      MovieAdvanceTimerKeyAndPostRedisplay();
      } // end lastKey == 13
      
   // Since input is no longer on the screen, go back to the default
   // object picking behavior (toggle solid/wire)
   ObjectPickedFunc = 0;
}

// While the user is inputting a new center of rotation,
// ObjectPickedFunc is set to this function.  Clicks on
// objects cause the center of rotation to be updated.
void NewCenterOfRotationPicked(void)
{
   int whichObject = lastPickedObject - 1;
   dispuiCenterOfRotation(objectCoord[whichObject][0],
                          objectCoord[whichObject][1],
                          objectCoord[whichObject][2]);
   MovieAdvanceTimerKeyAndPostRedisplay();
}

// dispui.c calls this function when the user presses <O>rigin
// or clicks on the origin box.
void displayCenterOfRotationRequested(void)
{
   static char prompt_string[100];

   // Build a prompt string
   sprintf(prompt_string,"Enter object number (1-9) or pick object to make new center of rotation");

   // Ask the user to enter an object
   KeyboardInputStart(prompt_string,
                     CenterOfRotationInputComplete);
                     
   // Setup this function pointer so that if the user clicks on
   // an object it becomes the new center of rotation.                     
   ObjectPickedFunc = NewCenterOfRotationPicked;
}

// You can use this function below to respond to clicks in the movie 
// Window without going into object selection mode.  This is especially useful
// if you have 2D text on the movie window, which might be removed or highlighted 
// when clicked.
int displayProcessRawMovieClick(int mouseX0,int mouseY0,int movieWindowWidth,int movieWindowHeight)
{
   return 0; // For now, do nothing
}

// When the user selects the staticMenu, dispui calls this function to let us
// know about it
void staticMenuCallback(int whichOption)
{
   // Only do something if new entry is something new
   if (statictype != whichOption) // nStaticFirst, nStaticAvg, nStaticNone
      {
      statictype = whichOption;
      MovieAdvanceTimerKeyAndPostRedisplay();
      }
}

void displayKeyboardFunc(int key)
{
   // Use this function to capture other keystrokes that
   // are not part of the basic user interface.  This function is
   // called by dispui when the keystrkoe command is not otherwise
   // recognized
}

//  This function is called by dispui.c when the user presses '?' or clicks on the help box.
//  You are welcome to change it anyway you like.  Note that DrawString is in dispui.c
void displayRenderHelpBox()
{
   int win_height = glutGet(GLUT_WINDOW_HEIGHT);
   int win_width = glutGet(GLUT_WINDOW_WIDTH);

   glColor3ub(0,0,255);
   DrawString(5,win_height-18,GLUT_BITMAP_HELVETICA_18,
         "CUJMOVIE Demo for Sept 2003 C/C++ Users Journal");

   glColor3ub(0,0,0);
   DrawString(5,win_height - 32,GLUT_BITMAP_HELVETICA_10,
   "ported by Chris Moth from work by Tim Callahan, Eric Swanson, and Terry Lybrand");

   glColor3ub(255,0,0);
   DrawString(5,win_height-48,GLUT_BITMAP_HELVETICA_12,"Journal of Molecular Graphics (1996) 14:39-41");

   glColor3ub(0,0,0);
   DrawString(10,win_height-80,GLUT_BITMAP_HELVETICA_12," Vanderbilt University, Nashville TN");
   DrawString(10,
         win_height-110,
         GLUT_BITMAP_HELVETICA_10,
         "http://www.structbio.vanderbilt.edu/~cmoth/mddisplay");

   glBegin(GL_LINE_STRIP);
      glVertex3i(0,win_height-52,0);
      glVertex3i(win_width-1,win_height-52,0);
   glEnd();


   // Display mouse examples help
   {
   static char* mouseExamples[][2] = 
   {  
      {"Left","Pick Objects or Controls"},
      {"Middle","Translate Scene"},
      {"Right","Rotate Scene"},
      {"Left + Right","Scale Scene"},
      {"",""},
      {"SHIFT+Left","Translate Scene"},

      {0,0} // <--- this 0 ends display loop
   };

   int example;
   int base_offset = 80;
   int right_offset = 275;

   glColor3ub(0,0,255);
   DrawString(win_width-right_offset,
               win_height-base_offset,
               GLUT_BITMAP_HELVETICA_12,
               "Mouse Button Functions");
   glColor3ub(0,0,0);
   glBegin(GL_LINE_STRIP);
      glVertex3i(win_width-right_offset,win_height-base_offset-2,0);
      glVertex3i(win_width-1,win_height-base_offset-2,0);
   glEnd();

   base_offset += 20;

   for (example=0;mouseExamples[example][0];example++)
      {
      glColor3ub(0,0,0);
      DrawString(win_width-right_offset,
               win_height-base_offset-example*15,
               GLUT_BITMAP_HELVETICA_12,
               mouseExamples[example][0]);
      glColor3ub(25,25,25);
      DrawString(win_width-right_offset+100,
                  win_height-base_offset-example*15,
                  GLUT_BITMAP_TIMES_ROMAN_10,
                  mouseExamples[example][1]);
      }
   } // End Mouse Examples

   // Key Character
   {
   static char* keyChars[][2] =
   {
      {"s","superimpose (smear)"},
      {"S","cancel smearing"},
      {"c",  "color palette"},
      {"0-9","solid colors"},
      {"r or 0","remove (color black)"},

      {"d","next display mode"},
      {"F1","Mono"},
      {"F2","Stereo Pairs"},
      {"F3","TRI"},
      {"F4","True Stereo (if supported)"},


      {0,0} // <--- this 0 ends display loop
   };

   int base_offset = 300;
   int row;

   glColor3ub(0,0,255);
   DrawString(2,win_height-base_offset,GLUT_BITMAP_HELVETICA_12,
      "Additional Keyboard Command Key Information");
   glColor3ub(0,0,0);
   glBegin(GL_LINE_STRIP);
      glVertex3i(0,win_height-base_offset-2,0);
      glVertex3i(win_width-1,win_height-base_offset-2,0);
   glEnd();

   glBegin(GL_LINE_STRIP);
      glVertex3i(4+200-12,win_height-base_offset-2,0);
      glVertex3i(4+200-12,0,0);
   glEnd();

   glBegin(GL_LINE_STRIP);
      glVertex3i(4+400-12,win_height-base_offset-2,0);
      glVertex3i(4+400-12,0,0);
   glEnd();

   base_offset += 20;

   for (row=0;keyChars[row*2][0];row++)
      {
      int column;
      for (column=0;column<2;column++)
         {
         int example = column*5 + row;
         static int offset[2] = {20,35};
         glColor3ub(0,0,0);
         DrawString(4+column*200,
               win_height-base_offset-row*15,GLUT_BITMAP_HELVETICA_10,keyChars[example][0]);
         glColor3ub(25,25,25);
         DrawString(column*200+offset[column],
               win_height-base_offset-row*15,GLUT_BITMAP_HELVETICA_10,keyChars[example][1]);
         }
      }
   }
}

int main(int argc,char** argv)
{
   
   // Apologies for the globals - but there are some setups 
   movie.frontClip = 1.0;
   movie.backClip =  2.0;
   movie.scaleFactor = 1.0;
   dispuiTranslation(0.0,0.0);
   movie.smear = 0;
   movie.imageView = nSingleImage;
   movie.hideLeftAndBottom = 0;
   
   // Our initial rotation matric will just be the identity matrix
   // (no rotation).  However, you can load alternate rotation
   // matrices - so the user can save their session and
   // resume it later (with a bit mode code - display.c does this
   dispuiInitRotationMatrix();

// Let's enable the quad stereo flag if an argument is typed
   glutInit(&argc,argv);
   // Initialize dispui with the size of the movie
   // in OpenGL world units (0,0,0) is at the center
   // 1 is the number of the first frame
   // argc > 1 means that if any argument supplied on the 
   // command line, enable quad buffered stereo
   dispuiInit(movieImageSize,1,argc > 1);


   // Let's go!
   dispuiMain("CUJ Demo Movie");
   return 0;
}
