//
// XML grammar implemented with Spirit (http://spirit.sourceforge.net/)
//
// Copyright (c) 2001, Daniel C. Nuffer, All rights reserved

#include <boost/spirit/spirit.hpp>

#ifdef USE_VECTOR
#include <vector>
#else
#include <boost/multi_pass.hpp>
#endif

#include <iostream>
#include <fstream>

using namespace spirit;
using namespace std;

typedef unsigned char char_t;

#ifdef USE_VECTOR
typedef vector<char_t>::iterator iterator_t;
#else
typedef boost::multi_pass<istream_iterator<char_t> > iterator_t;
#endif

typedef rule<iterator_t> rule_t;

static void
Parse(rule_t const& rule, char const* filename)
{
    cout << "/////////////////////////////////////////////////////////\n\n";

    ifstream in(filename);
    if (!in)
    {
        cout << "Could not open file: " << filename << endl;
        return;
    }

    // Turn of white space skipping on the stream
    in.unsetf(ios::skipws);

#ifdef USE_VECTOR
    iterator_t first = data.begin();
    iterator_t last = data.end();
#else
    iterator_t first = boost::make_multi_pass(istream_iterator<char_t>(in));
    iterator_t last = boost::make_multi_pass(istream_iterator<char_t>());
#endif
    match m = rule.parse(first, last);


    if (m && first == last)
    {
        cout << "\t\t" << filename << " Parses OK\n\n\n";
    }
    else
    {
      cout << "\t\t" << filename << " Fails Parsing\n" << endl;
      for (int i = 0; i < 50; ++i)
      {
          cout << *first++;
      }
      cout << endl;
    }
}

int
main(int argc, char* argv[])
{

    // Start grammar definition
    rule_t    prolog, element, Misc, Letter, Digit, CombiningChar, 
        Extender, Reference,
        CData, doctypedecl, XMLDecl, SDDecl, VersionInfo, EncodingDecl,
        VersionNum, Eq, EmptyElemTag, STag, content, ETag, Attribute,
        AttValue, CharDataChar, CharData, Comment, CDSect,
        CharRef, EntityRef, EncName, BaseChar, document, Char, S, NameChar, 
        Name, Comment1;

    chlit<>    Ampersand('&');
    chlit<>    DblQuote('"');
    chlit<>    Quote('\'');
    chlit<>    LessThan('<');

    chlit<>    x9('\x9');
    chlit<>    xA('\xA');
    chlit<>    xD('\xD');
    range<char_t>    x20xFF('\x20', '\xFF');
    chlit<>    x20('\x20');

    // This is the start rule for XML
    document =         prolog >> element >> *Misc;
    Char =             x9 | xA | xD | x20xFF;

    S =                 +(x20 | x9 | xD | xA);

    NameChar =         Letter | Digit | '.' | '-' | '_' | ':' | CombiningChar | Extender;
    Name =             (Letter | '_' | ':') >> *(NameChar);

    AttValue =         '"' >> *((anychar - (LessThan | Ampersand | DblQuote)) | Reference)
                            >> '"'
                        | '\'' >> *((anychar - (LessThan | Ampersand | Quote)) | Reference)
                            >> '\'';

    CharDataChar = anychar - (LessThan | Ampersand);
    CharData =        *(CharDataChar - str_p("]]>"));

    Comment1 = *((Char - ch_p('-')) | (ch_p('-') >> (Char - ch_p('-'))));
    Comment =        str_p("<!--") >> Comment1 >> str_p("-->");

    CDSect =            str_p("<![CDATA[") >> CData >> str_p("]]>");
    CData =            *(Char - str_p("]]>"));

    prolog =            !XMLDecl >> *Misc >> !(doctypedecl >> *Misc);
    XMLDecl =        str_p("<?xml") >> VersionInfo >> !EncodingDecl >> !SDDecl >> !S >> str_p("?>");
    VersionInfo =    S >> str_p("version") >> Eq >> (Quote >> VersionNum >> Quote
                            | DblQuote >> VersionNum >> DblQuote);
    Eq =                !S >> '=' >> !S;
    VersionNum =    +(alnum | '_' | '.' | ':' | '-');
    Misc =            Comment | S;

    doctypedecl =    str_p("<!DOCTYPE") >> *(Char - (chlit<>('[') | '>')) >> !('[' >> *(Char - ']') >> ']') >> '>';
    SDDecl =            S >> str_p("standalone") >> Eq >> ((Quote >> (str_p("yes") | str_p("no")) >> Quote)
                            | (DblQuote >> (str_p("yes") | str_p("no")) >> DblQuote));

    element =        (EmptyElemTag
                        | STag >> content >> ETag);

    STag =            '<' >> Name >> *(S >> Attribute) >> !S >> '>';
    Attribute =        Name >> Eq >> AttValue;
    ETag =            str_p("</") >> Name >> !S >> '>';
    content =        !CharData >> *((Reference | CDSect | Comment | element) >> !CharData);
    EmptyElemTag =    '<' >> Name >> *(S >> Attribute) >> !S >> str_p("/>");

    CharRef =        str_p("&#") >> +digit >> ';'
                        | str_p("&#x") >> +xdigit >> ';';

    Reference =        EntityRef | CharRef;
    EntityRef =        '&' >> Name >> ';';

    EncodingDecl =    S >> str_p("encoding") >> Eq >> (DblQuote >> EncName >> DblQuote |
                            Quote >> EncName >> Quote);
    EncName =        alpha >> *(alnum | '.' | '_' | '-');


    Letter =         BaseChar;// | Ideographic;
    BaseChar =        range<char_t>('\x41', '\x5A') | range<char_t>('\x61', '\x7A')
                        | range<char_t>('\xC0', '\xD6') | range<char_t>('\xD8', '\xF6')
                        | range<char_t>('\xF8', '\xFF');

    // Ideographic =
    CombiningChar =    nothing;
    Digit =            digit;
    Extender =        chlit<char_t>('\xB7');

    if (argc == 2)
    {
        Parse(document, argv[1]);
    }
    else
    {
        cout << "no filename given" << endl;
    }

    return 0;
}
